@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  *
  * Authors:
  *      "Daniel Kopecek" <dkopecek@redhat.com>
@@ -32,22 +32,20 @@
 #include <sys/stat.h>
 #include <limits.h>
 #include <errno.h>
-#include <assume.h>
 #include <pcre.h>
-#include <libgen.h>
 
+#include "oscap_helpers.h"
 #include "fsdev.h"
 #include "_probe-api.h"
 #include "probe/entcmp.h"
-#include "alloc.h"
 #include "debug_priv.h"
 #include "oval_fts.h"
-#if defined(__SVR4) && defined(__sun)
+#if defined(OS_SOLARIS)
 #include "fts_sun.h"
 #include <sys/mntent.h>
 #include <libzonecfg.h>
 #include <sys/avl.h>
-#elif defined(_AIX)
+#elif defined(OS_AIX)
 #include "fts_sun.h"
 #else
 #include <fts.h>
@@ -57,10 +55,7 @@
 
 static OVAL_FTS *OVAL_FTS_new()
 {
-	OVAL_FTS *ofts;
-
-	ofts = oscap_talloc(OVAL_FTS);
-	memset(ofts, 0, sizeof(*ofts));
+	OVAL_FTS *ofts = calloc(1, sizeof(OVAL_FTS));
 
 	ofts->max_depth  = -1;
 	ofts->direction  = -1;
@@ -97,9 +92,7 @@ static int pathlen_from_ftse(int fts_pathlen, int fts_namelen)
 
 static OVAL_FTSENT *OVAL_FTSENT_new(OVAL_FTS *ofts, FTSENT *fts_ent)
 {
-	OVAL_FTSENT *ofts_ent;
-
-	ofts_ent = oscap_talloc(OVAL_FTSENT);
+	OVAL_FTSENT *ofts_ent = calloc(1, sizeof(OVAL_FTSENT));
 
 	ofts_ent->fts_info = fts_ent->fts_info;
 	/* The 'shift' variable stores length of the prefix if the prefix
@@ -149,7 +142,7 @@ static void OVAL_FTSENT_free(OVAL_FTSENT *ofts_ent)
 	return;
 }
 
-#if defined(__SVR4) && defined(__sun)
+#if defined(OS_SOLARIS)
 #ifndef MNTTYPE_SMB
 #define MNTTYPE_SMB	"smb"
 #endif
@@ -236,7 +229,7 @@ int load_zones_path_list()
 				    name);
 				continue;
 			}
-			if (realpath(rpath, temp->zpath) != NULL)
+			if (oscap_realpath(rpath, temp->zpath) != NULL)
 				avl_add(&avl_tree_list, temp);
 		}
 	}
@@ -272,7 +265,7 @@ static bool valid_local_zone(const char *path)
 
 static bool OVAL_FTS_localp(OVAL_FTS *ofts, const char *path, void *id)
 {
-#if defined(__SVR4) && defined(__sun)
+#if defined(OS_SOLARIS)
 	if (id != NULL && (*(char*)id) != '\0') {
 		/* if not a valid local fs skip */
 		if (valid_local_fs((char*)id)) {
@@ -396,7 +389,7 @@ static int badpartial_check_slash(const char *pattern)
 	if (regex == NULL) {
 		dE("Failed to validate the pattern: pcre_compile(): "
 		   "error: '%s', error offset: %d, pattern: '%s'.\n",
-		   errofs, errptr, pattern);
+		   errptr, errofs, pattern);
 		return -1;
 	}
 	ret = pcre_fullinfo(regex, NULL, PCRE_INFO_FIRSTBYTE, &fb);
@@ -569,7 +562,7 @@ static int process_pattern_match(const char *path, pcre **regex_out)
 		   OVAL definitions that use ".*" as 'path' and then
 		   uncomment this.
 
-		dI("pcre_exec() returned PCRE_ERROR_PARTIAL for pattern '%s' "
+		dD("pcre_exec() returned PCRE_ERROR_PARTIAL for pattern '%s' "
 		   "and test path '%s'.\n", pattern, test_path1);
 		ret = pcre_exec(regex, NULL, test_path2, strlen(test_path2),
 			0, PCRE_PARTIAL, NULL, 0);
@@ -585,7 +578,7 @@ static int process_pattern_match(const char *path, pcre **regex_out)
 		*/
 		break;
 	case PCRE_ERROR_BADPARTIAL:
-		dI("pcre_exec() returned PCRE_ERROR_BADPARTIAL for pattern "
+		dD("pcre_exec() returned PCRE_ERROR_BADPARTIAL for pattern "
 		   "'%s' and a test path '%s'. Falling back to "
 		   "pcre_fullinfo().\n", pattern, test_path1);
 		pcre_free(regex);
@@ -650,9 +643,9 @@ static int process_pattern_match(const char *path, pcre **regex_out)
 	}
 
 	if (regex == NULL) {
-		dI("Disabling partial match optimization.");
+		dD("Disabling partial match optimization.");
 	} else {
-		dI("Enabling partial match optimization using "
+		dD("Enabling partial match optimization using "
 		   "pattern: '%s'.", pattern);
 		if (regex_out != NULL)
 			*regex_out = regex;
@@ -695,9 +688,13 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 	pcre *regex = NULL;
 	struct stat st;
 
-	assume_d((path == NULL && filename == NULL && filepath != NULL)
-		 || (path != NULL && filepath == NULL), NULL);
-	assume_d(behaviors != NULL, NULL);
+	if ((path != NULL || filename != NULL || filepath == NULL)
+			&& (path == NULL || filepath != NULL)) {
+		return NULL;
+	}
+	if (behaviors == NULL) {
+		return NULL;
+	}
 
 	if (path)
 		PROBE_ENT_AREF(path, r0, "operation", /**/);
@@ -711,7 +708,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 		path_op = OVAL_OPERATION_EQUALS;
 	}
 #if defined(OSCAP_FTS_DEBUG)
-	dI("path_op: %u, '%s'.", path_op, oval_operation_get_text(path_op));
+	dD("path_op: %u, '%s'.", path_op, oval_operation_get_text(path_op));
 #endif
 	if (path) { /* filepath == NULL */
 		PROBE_ENT_STRVAL(path, cstr_path, sizeof cstr_path,
@@ -739,7 +736,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 		return (NULL);
 	}
 #if defined(OSCAP_FTS_DEBUG)
-	dI("bh.max_depth: %s => max_depth: %d", cstr_buff, max_depth);
+	dD("bh.max_depth: %s => max_depth: %d", cstr_buff, max_depth);
 #endif
 	SEXP_free(r0);
 
@@ -759,7 +756,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 		return (NULL);
 	}
 #if defined(OSCAP_FTS_DEBUG)
-	dI("bh.direction: %s => direction: %d", cstr_buff, direction);
+	dD("bh.direction: %s => direction: %d", cstr_buff, direction);
 #endif
 	SEXP_free(r0);
 
@@ -785,7 +782,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 		recurse = OVAL_RECURSE_SYMLINKS_AND_DIRS;
 	}
 #if defined(OSCAP_FTS_DEBUG)
-	dI("bh.recurse: %s => recurse: %d", cstr_buff, recurse);
+	dD("bh.recurse: %s => recurse: %d", cstr_buff, recurse);
 #endif
 	SEXP_free(r0);
 
@@ -811,7 +808,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 		filesystem = OVAL_RECURSE_FS_ALL;
 	}
 #if defined(OSCAP_FTS_DEBUG)
-	dI("bh.filesystem: %s => filesystem: %d", cstr_buff, filesystem);
+	dD("bh.filesystem: %s => filesystem: %d", cstr_buff, filesystem);
 #endif
 	SEXP_free(r0);
 
@@ -828,7 +825,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 		if (process_pattern_match(cstr_path, &regex) != 0)
 			return NULL;
 		paths[0] = extract_fixed_path_prefix(cstr_path);
-		dI("Extracted fixed path: '%s'.", paths[0]);
+		dD("Extracted fixed path: '%s'.", paths[0]);
 	} else {
 		paths[0] = strdup("/");
 	}
@@ -875,10 +872,10 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 	}
 
 	if (filesystem == OVAL_RECURSE_FS_LOCAL) {
-#if   defined(__SVR4) && defined(__sun)
+#if defined(OS_SOLARIS)
 		ofts->localdevs = NULL;
 #else
-		ofts->localdevs = fsdev_init(NULL, 0);
+		ofts->localdevs = fsdev_init();
 		if (ofts->localdevs == NULL) {
 			dE("fsdev_init() failed.");
 			/* One dummy read to get rid of an uninitialized
@@ -914,7 +911,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 		ofts->ofts_sfilepath = SEXP_ref(filepath);
 	}
 
-#if defined(__SVR4) && defined(__sun)
+#if defined(OS_SOLARIS)
 	if (load_zones_path_list() != 0) {
 		dE("Failed to load zones path info. Recursing non-global zones.");
 		free_zones_path_list();
@@ -927,7 +924,7 @@ OVAL_FTS *oval_fts_open_prefixed(const char *prefix, SEXP_t *path, SEXP_t *filen
 }
 
 static inline int _oval_fts_is_local(OVAL_FTS *ofts, FTSENT *fts_ent) {
-# if defined (__SVR4) && defined(__sun)
+# if defined(OS_SOLARIS)
 	/* pseudo filesystems will be skipped */
 	/* don't recurse into remote fs if local is specified */
 	return ((fts_ent->fts_info == FTS_D || fts_ent->fts_info == FTS_SL)
@@ -966,7 +963,7 @@ static FTSENT *oval_fts_read_match_path(OVAL_FTS *ofts)
 		}
 
 #if defined(OSCAP_FTS_DEBUG)
-		dI("fts_path: '%s' (l=%d)."
+		dD("fts_path: '%s' (l=%d)."
 		   "fts_name: '%s' (l=%d).\n"
 		   "fts_info: %u.\n", fts_ent->fts_path, fts_ent->fts_pathlen,
 		   fts_ent->fts_name, fts_ent->fts_namelen, fts_ent->fts_info);
@@ -974,7 +971,7 @@ static FTSENT *oval_fts_read_match_path(OVAL_FTS *ofts)
 
 		if (fts_ent->fts_info == FTS_SL) {
 #if defined(OSCAP_FTS_DEBUG)
-			dI("Only the target of a symlink gets reported, skipping '%s'.", fts_ent->fts_path, fts_ent->fts_name);
+			dD("Only the target of a symlink gets reported, skipping '%s'.", fts_ent->fts_path, fts_ent->fts_name);
 #endif
 			fts_set(ofts->ofts_match_path_fts, fts_ent, FTS_FOLLOW);
 			continue;
@@ -992,12 +989,13 @@ static FTSENT *oval_fts_read_match_path(OVAL_FTS *ofts)
 			continue;
 		}
 
+		const size_t shift = ofts->prefix ? strlen(ofts->prefix) : 0;
 		/* partial match optimization for OVAL_OPERATION_PATTERN_MATCH operation on path and filepath */
 		if (ofts->ofts_path_regex != NULL && fts_ent->fts_info == FTS_D) {
 			int ret, svec[3];
 
 			ret = pcre_exec(ofts->ofts_path_regex, ofts->ofts_path_regex_extra,
-					fts_ent->fts_path, fts_ent->fts_pathlen, 0, PCRE_PARTIAL,
+					fts_ent->fts_path+shift, fts_ent->fts_pathlen-shift, 0, PCRE_PARTIAL,
 					svec, sizeof(svec) / sizeof(svec[0]));
 			if (ret < 0) {
 				switch (ret) {
@@ -1019,7 +1017,6 @@ static FTSENT *oval_fts_read_match_path(OVAL_FTS *ofts)
 		    || (!ofts->ofts_sfilepath && fts_ent->fts_info != FTS_D))
 			continue;
 
-		const size_t shift = ofts->prefix ? strlen(ofts->prefix) : 0;
 		stmp = SEXP_string_newf("%s", fts_ent->fts_path + shift);
 
 		if (ofts->ofts_sfilepath)
@@ -1032,6 +1029,15 @@ static FTSENT *oval_fts_read_match_path(OVAL_FTS *ofts)
 
 		if (ores == OVAL_RESULT_TRUE)
 			break;
+		if (ofts->ofts_path_op == OVAL_OPERATION_EQUALS) {
+			/* At this point the comparison result isn't OVAL_RESULT_TRUE. Since
+			we passed the exact path (from filepath or path elements) to
+			fts_open() we surely know that we can't find other items that would
+			be equal. Therefore we can terminate the matching. This can happen
+			if the filepath or path element references a variable that has
+			multiple different values. */
+			return NULL;
+		}
 	} /* for (;;) */
 
 	/*
@@ -1074,7 +1080,7 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 			char * const paths[2] = { ofts->ofts_match_path_fts_ent->fts_path, NULL };
 
 #if defined(OSCAP_FTS_DEBUG)
-			dI("fts_open args: path: \"%s\", options: %d.",
+			dD("fts_open args: path: \"%s\", options: %d.",
 				paths[0], ofts->ofts_recurse_path_fts_opts);
 #endif
 			/* reset errno as fts_open() doesn't do it itself. */
@@ -1121,7 +1127,7 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 			}
 
 #if defined(OSCAP_FTS_DEBUG)
-			dI("fts_path: '%s' (l=%d)."
+			dD("fts_path: '%s' (l=%d)."
 			   "fts_name: '%s' (l=%d).\n"
 			   "fts_info: %u.\n", fts_ent->fts_path, fts_ent->fts_pathlen,
 			   fts_ent->fts_name, fts_ent->fts_namelen, fts_ent->fts_info);
@@ -1166,10 +1172,11 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 				/* limit recursion only to selected file types */
 				switch (fts_ent->fts_info) {
 				case FTS_D:
-					if (!(ofts->recurse & OVAL_RECURSE_DIRS)) {
+					if (!(ofts->recurse & OVAL_RECURSE_DIRS) && !(ofts->recurse & OVAL_RECURSE_SYMLINKS && ofts->following)) {
 						fts_set(ofts->ofts_recurse_path_fts, fts_ent, FTS_SKIP);
 						continue;
 					}
+					ofts->following = 0;
 					break;
 				case FTS_SL:
 					if (!(ofts->recurse & OVAL_RECURSE_SYMLINKS)) {
@@ -1177,6 +1184,7 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 						continue;
 					}
 					fts_set(ofts->ofts_recurse_path_fts, fts_ent, FTS_FOLLOW);
+					ofts->following = 1;
 					break;
 				default:
 					continue;
@@ -1209,7 +1217,7 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 				char * const paths[2] = { ofts->ofts_recurse_path_curpth, NULL };
 
 #if defined(OSCAP_FTS_DEBUG)
-				dI("fts_open args: path: \"%s\", options: %d.",
+				dD("fts_open args: path: \"%s\", options: %d.",
 					paths[0], ofts->ofts_recurse_path_fts_opts);
 #endif
 				/* reset errno as fts_open() doesn't do it itself. */
@@ -1251,7 +1259,7 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 				if (ofts->ofts_recurse_path_curdepth == 0)
 					ofts->ofts_recurse_path_devid = fts_ent->fts_statp->st_dev;
 				*/
-#if   defined(__SVR4) && defined(__sun)
+#if defined(OS_SOLARIS)
 				if ((!OVAL_FTS_localp(ofts, fts_ent->fts_path,
 				    (fts_ent->fts_statp != NULL) ?
 				    &fts_ent->fts_statp->st_fstype : NULL)))
@@ -1276,7 +1284,7 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 						break;
 					}
 				} else {
-					if (fts_ent->fts_info != FTS_D) {
+					if (fts_ent->fts_info != FTS_D && fts_ent->fts_info != FTS_DP && fts_ent->fts_info != FTS_DC) {
 						SEXP_t *stmp;
 
 						stmp = SEXP_string_newf("%s", fts_ent->fts_name);
@@ -1302,7 +1310,7 @@ static FTSENT *oval_fts_read_recurse_path(OVAL_FTS *ofts)
 			if (!strcmp(ofts->ofts_recurse_path_curpth, "/"))
 				break;
 
-			ofts->ofts_recurse_path_curpth = dirname(ofts->ofts_recurse_path_curpth);
+			ofts->ofts_recurse_path_curpth = oscap_dirname(ofts->ofts_recurse_path_curpth);
 			ofts->ofts_recurse_path_curdepth++;
 		}
 
@@ -1322,7 +1330,7 @@ OVAL_FTSENT *oval_fts_read(OVAL_FTS *ofts)
 	FTSENT *fts_ent;
 
 #if defined(OSCAP_FTS_DEBUG)
-	dI("ofts: %p.", ofts);
+	dD("ofts: %p.", ofts);
 #endif
 
 	if (ofts == NULL)
@@ -1338,6 +1346,12 @@ OVAL_FTSENT *oval_fts_read(OVAL_FTS *ofts)
 		if (ofts->ofts_sfilepath) {
 			fts_ent = ofts->ofts_match_path_fts_ent;
 			ofts->ofts_match_path_fts_ent = NULL;
+			if (ofts->filesystem == OVAL_RECURSE_FS_LOCAL
+				&& (!OVAL_FTS_localp(ofts, fts_ent->fts_path,
+					(fts_ent->fts_statp != NULL) ?
+					&fts_ent->fts_statp->st_dev : NULL))) {
+				continue;
+			}
 			break;
 		} else {
 			fts_ent = oval_fts_read_recurse_path(ofts);
@@ -1381,7 +1395,7 @@ int oval_fts_close(OVAL_FTS *ofts)
 	fsdev_free(ofts->localdevs);
 
 	OVAL_FTS_free(ofts);
-#if defined(__SVR4) && defined(__sun)
+#if defined(OS_SOLARIS)
 	free_zones_path_list();
 #endif
 
